/*
 * Copyright (c) 2009 The openGion Project.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 * either express or implied. See the License for the specific language
 * governing permissions and limitations under the License.
 */
package org.opengion.hayabusa.resource;

import org.opengion.hayabusa.common.HybsSystem;
import org.opengion.hayabusa.common.HybsSystemException;

import org.opengion.fukurou.util.ApplicationInfo;
import org.opengion.fukurou.db.DBUtil;

import java.util.Map;
import java.util.HashMap;
import java.util.LinkedHashMap;

/**
 * ユーザーアクセス画面管理テーブルを維持する為のクラスです。
 * @og.group リソース管理
 *
 * @version  4.1.1.0
 * @author   Sen.Li
 * @since    JDK5.0,
 */
public final class UserAccessTable {
	// 5.3.0.0 (2010/12/01) 画面アクセスの履歴(順番)を管理する機能を追加
//	private final static String GEA09_QUERY  = "SELECT GUIKEY,SEQNO,NAME_JA,LAST_ACCESS,FGFAVORITE,CLASSIFY FROM GEA09"
//									+ " WHERE SYSTEM_ID=? AND USERID=?"
//									+ " ORDER BY SEQNO,CLASSIFY,GUIKEY";
	private final static String GEA09_QUERY  = "SELECT GUIKEY,SEQNO,NAME_JA,LAST_ACCESS,FGFAVORITE,CLASSIFY,NEXTGUI FROM GEA09"
									+ " WHERE SYSTEM_ID=? AND USERID=?"
									+ " ORDER BY SEQNO,CLASSIFY,GUIKEY";
	private final static String GEA09_DELETE = "DELETE FROM GEA09"
									+ " WHERE SYSTEM_ID=? AND USERID=? AND GUIKEY=?";
	private final static String GEA09_INSERT = "INSERT INTO GEA09(SYSTEM_ID,USERID,GUIKEY,SEQNO,NAME_JA,CLASSIFY,FGJ"
									+ ",DYSET,USRSET,PGUPD) VALUES(?,?,?,?,?,?,?,?,?,?)";
	// 5.3.0.0 (2010/12/01) 画面アクセスの履歴(順番)を管理する機能を追加
//	private final static String GEA09_UPDATE = "UPDATE GEA09 SET LAST_ACCESS=?"
//									+ " WHERE SYSTEM_ID=? AND USERID=? AND GUIKEY=?";
	private final static String GEA09_UPDATE = "UPDATE GEA09 SET LAST_ACCESS=?,NEXTGUI=?"
									+ " WHERE SYSTEM_ID=? AND USERID=? AND GUIKEY=?";


	private final static String DBID = HybsSystem.sys( "RESOURCE_DBID" );
	private final static int GEA09_GUIKEY		= 0;
	private final static int GEA09_SEQNO		= 1;
	private final static int GEA09_NAME_JA		= 2;
	private final static int GEA09_LAST_ACCESS	= 3;
	private final static int GEA09_FGFAVORITE	= 4;
	private final static int GEA09_CLASSIFY		= 5;
	private final static int GEA09_NEXTGUI		= 6;	// 5.3.0.0 (2010/12/01) 追加
	private final static String  FGFAVORITE_ON	= "1";
	private static enum FgKeys { GUIMAP_ONLY,GEA09_ONLY };

	/**
	 * コンストラクター
	 * オブジェクトを作成できないように、privateにします。
	 *
	 * @og.rev 4.3.1.1 (2008/08/23) コンストラクタは、void 宣言を付けません。
	 */
//	private void UserAccessTable() { }
	private UserAccessTable() { }

	/**
	 * ユーザーアクセス画面管理テーブルを整理します。
	 * このメソッドでは guiMap へのセットをしています（副作用の扱い）。
	 *
	 * @og.rev 4.1.1.0 (2008/01/30) 新規追加
	 * @og.rev 5.3.0.0 (2010/12/01) 画面アクセスの履歴(順番)を管理する機能を追加
	 *
	 * @param  guiMap         Map<String,GUIInfo> 画面オブジェクトマップへの参照
	 * @param  systemId       String システムID
	 * @param  userId         String ユーザーID
	 * @param  lang           String 言語
	 * @return Map<String,FavoriteGUIData> お気に入りマップ
	 */
	public static Map<String,FavoriteGUIData> makeAccessDB( 
							final Map<String,GUIInfo> guiMap,final String systemId,final String userId,final String lang ) {
		String today			= HybsSystem.getDate( "yyyyMMddHHmmss" );
		Map<String,FgKeys>	conGuiKey = new HashMap<String,FgKeys>();
		/** コネクションにアプリケーション情報を追記するかどうか指定 */
		boolean USE_DB_APPLICATION_INFO  = HybsSystem.sysBool( "USE_DB_APPLICATION_INFO" ) ;
		ApplicationInfo appInfo = null;
		if( USE_DB_APPLICATION_INFO ) {
			appInfo = new ApplicationInfo();
			// ユーザーID,IPアドレス,ホスト名
			appInfo.setClientInfo( systemId,HybsSystem.HOST_ADRS,HybsSystem.HOST_NAME );
			// 画面ID,操作,プログラムID
			appInfo.setModuleInfo( "UserAccessTable",userId,"makeAccessDB" );
		}

		// guiMapのキーで、新マップを作成、
		for( String conKey : guiMap.keySet() ) {
			conGuiKey.put( conKey,FgKeys.GUIMAP_ONLY );
		}

		Map<String,FavoriteGUIData> favoriteGuiMap = new LinkedHashMap<String,FavoriteGUIData>();
		// 4.1.1.0（2008/01/22）ユーザーアクセスの日付を取得し、accessPastDays 前の日付を計算する。
		String accessPastDays	= HybsSystem.sys( "ACCESS_TOKEI_PAST_DAYS" );
		int    diffDate			= ( accessPastDays == null ) ? 0 : -Integer.parseInt( accessPastDays );
		String judgeTime		= HybsSystem.getDate( today.substring( 0,8 ),diffDate );
		String lastAccessTime	= null;
		String key 				= null;
		String[]   args			= new String[] { systemId,userId };
		String[][] vals			= DBUtil.dbExecute( GEA09_QUERY,args,appInfo,DBID );
		int len					= vals.length;

		for( int i=0; i<len; i++ ) {
			key = vals[i][GEA09_GUIKEY];
			GUIInfo gui = guiMap.get( key );
			if( gui == null ) {
				conGuiKey.put( key,FgKeys.GEA09_ONLY );
			}
			else {
				// 5.3.0.0 (2010/12/01) 画面アクセスの履歴(順番)を管理する機能を追加
				String nxtgui = vals[i][GEA09_NEXTGUI];
				if( nxtgui != null ) {
					String[] keys = nxtgui.split( "," );
					for( String nextKey : keys ) {
						gui.setNextGuiKey( nextKey );		// DB から復活
					}
				}

				conGuiKey.remove( key );
				// 4.1.1.0（2008/01/22）お気に入りマップの作成
				if( FGFAVORITE_ON.equals( vals[i][GEA09_FGFAVORITE] ) && gui.isRead() ) {
					favoriteGuiMap.put( key,new FavoriteGUIData( vals[i][GEA09_GUIKEY],vals[i][GEA09_SEQNO]
						,vals[i][GEA09_NAME_JA],vals[i][GEA09_CLASSIFY] ) );
				}
				// 4.1.1.0（2008/01/22）ACCESS_TOKEI_PAST_DAYSに定義された期間以内の画面を格上する。
				lastAccessTime = vals[i][GEA09_LAST_ACCESS];
				if( lastAccessTime.length() >= 8 ) {
					if( lastAccessTime.compareTo( judgeTime ) >=0 ) {
						gui.setLevelUp();
					}
				}
			}
		}

		ResourceManager resource = ResourceFactory.newInstance( lang );
		String undefined = resource.getLabel( "UNDEFINED" );
		FgKeys fgKey = null;
		for( String conKey : conGuiKey.keySet() ) {
			fgKey = conGuiKey.get( conKey );
			switch ( fgKey ) {
				case GUIMAP_ONLY:
					args = new String[] { systemId,userId,conKey,"9999999",guiMap.get(conKey).getLabel()
										  ,undefined,"1",today,userId,"UsrAccsTbl" };
	                DBUtil.dbExecute( GEA09_INSERT,args,appInfo,DBID );
					break;
				case GEA09_ONLY:
					args = new String[] { systemId,userId,conKey };
					DBUtil.dbExecute( GEA09_DELETE,args,appInfo,DBID );
					break;
				default:
					String errMsg = "guiMapとGEA09の突合せフラグが GUIMAP_ONLY と GEA09_ONLY 以外の値がセットされています。"
								  + "画面ID：" + conKey + " フラグ：" + fgKey; 
					throw new HybsSystemException( errMsg );
			}
		}
		return favoriteGuiMap;
	}

	/**
	 * ユーザーアクセス画面管理テーブルの最終アクセス時間を更新します。
	 *
	 * @og.rev 4.1.1.0 (2008/01/30) 新規追加
	 * @og.rev 5.3.0.0 (2010/12/01) 画面アクセスの履歴(順番)を管理する機能を追加
	 *
	 * @param systemId       String システムID
	 * @param userId         String ユーザーID
	 * @param guiKey         String 画面ID
	 * @param lastAccessTime String 画面の最終アクセス時間
	 * @param nextGuiKeys    String 次にアクセスしている画面IDのCSV文字列
	 */
//	public static void updateLastAccessTime( final String systemId,final String userId,final String guiKey,final String lastAccessTime ) {
	public static void updateLastAccessTime( final String systemId,final String userId,final String guiKey,final String lastAccessTime,final String nextGuiKeys ) {
		/** コネクションにアプリケーション情報を追記するかどうか指定 */
		boolean USE_DB_APPLICATION_INFO  = HybsSystem.sysBool( "USE_DB_APPLICATION_INFO" ) ;
		ApplicationInfo appInfo = null;
		if( USE_DB_APPLICATION_INFO ) {
			appInfo = new ApplicationInfo();
			// ユーザーID,IPアドレス,ホスト名
			appInfo.setClientInfo( systemId,HybsSystem.HOST_ADRS,HybsSystem.HOST_NAME );
			// 画面ID,操作,プログラムID
			appInfo.setModuleInfo( "UserAccessTable",userId,"updateLastAccessTime" );
		}

		// 5.3.0.0 (2010/12/01) 画面アクセスの履歴(順番)を管理する機能を追加
//		String[] args = new String[] { lastAccessTime,systemId,userId,guiKey };
		String[] args = new String[] { lastAccessTime,nextGuiKeys,systemId,userId,guiKey };
		DBUtil.dbExecute( GEA09_UPDATE,args,appInfo,DBID );
	}
}
